from pymol.Qt import QtCore, QtWidgets


class ResizableMessageBox(QtWidgets.QMessageBox):

    _EVENT_TYPES = (
            QtCore.QEvent.UpdateRequest,
            QtCore.QEvent.WinIdChange,
            QtCore.QEvent.ShowToParent,
            )

    _UNWANTED_WINDOW_FLAGS = (
            QtCore.Qt.MSWindowsFixedSizeDialogHint |
            0)

    def _make_resizable(self):
        textEdit = self.findChild(QtWidgets.QTextEdit)
        if textEdit is None:
            return

        self.setSizeGripEnabled(True)

        ex = QtWidgets.QSizePolicy.Expanding
        for w in [self, textEdit]:
            w.setMaximumSize(0xffffff, 0xffffff)
            w.setSizePolicy(ex, ex)

    def event(self, e):
        if e.type() in self._EVENT_TYPES:
            self._make_resizable()

        return super(ResizableMessageBox, self).event(e)



from enum import IntEnum
from dataclasses import dataclass

from pymol.Qt import QtGui, QtWidgets
from pymol.Qt import QtCore
Qt = QtCore.Qt
QSI = QtGui.QStandardItem  # For brevity


class SceneDictIndex(IntEnum):
    QPIXMAP = 0
    MESSAGE = 1
    ACTIONS = 2


class SceneTableColumn(IntEnum):
    NAME = 0
    IMAGE = 1
    MESSAGE = 2
    ACTIONS = 3


@dataclass
class SceneWithTablePosition:
    name: str
    position: int


class ScenePanel(QtWidgets.QWidget):
    '''
    Scene Panel dialog for displaying and altering scenes in PyMOL.
    '''

    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self)
        self.cmd = parent.cmd
        self.scene_list = self._update_scene_list()
        # EXAMPLE: scene_dict = {name:[QPixmap,Message,Actions]}
        self.scene_dict = {}

        self._build_table_elements(parent)
        self._populate_data()
        self._set_button_connections()

        self.resize(365, 700)

    def _build_table_elements(self, parent):
        '''
        Create the various elements in the table and add them to the
        proper layout.
        '''
        self.setWindowTitle('Scene Panel')
        layout = QtWidgets.QVBoxLayout(self)
        self.setLayout(layout)

        top_layout = QtWidgets.QGridLayout()
        layout.addLayout(top_layout)
        mid_layout = QtWidgets.QGridLayout()
        layout.addLayout(mid_layout)
        low_layout = QtWidgets.QGridLayout()
        layout.addLayout(low_layout)

        # Top Elements
        self.instructionLabel = QtWidgets.QLabel(self)
        self.instructionLabel.setText(
            'Double click selected thumbnail to \nload into Workspace.')
        top_layout.addWidget(self.instructionLabel, 0, 0)

        self.addSceneButton = QtWidgets.QPushButton(self)
        self.addSceneButton.setText('Add Scene')
        top_layout.addWidget(self.addSceneButton, 0, 1)

        # Mid Elements
        self.sceneTableWidget = QtWidgets.QTableWidget(self)
        mid_layout.addWidget(self.sceneTableWidget, 0, 0)
        self.sceneTableWidget.viewport().installEventFilter(self)

        self.sceneTableWidget.itemChanged.connect(self._item_changed)
        self.sceneTableWidget.selectionModel().selectionChanged.connect(
            self._selection_changed)
        self.sceneTableWidget.setSelectionBehavior(
            QtWidgets.QAbstractItemView.SelectRows)
        self.sceneTableWidget.verticalHeader().setSectionsMovable(True)

        # Lower Buttom Elements
        self.deleteButton = QtWidgets.QPushButton(self)
        self.deleteButton.setText("Delete Scene")
        low_layout.addWidget(self.deleteButton, 0, 1)
        self.deleteButton.setEnabled(False)

        self.updateButton = QtWidgets.QPushButton(self)
        self.updateButton.setText("Update Scene")
        low_layout.addWidget(self.updateButton, 0, 0)
        self.updateButton.setEnabled(False)

    def _set_button_connections(self):
        self.addSceneButton.clicked.connect(self._add_scene)
        self.deleteButton.clicked.connect(self._delete_scene)
        self.updateButton.clicked.connect(self._update_scene)
        self.sceneTableWidget.doubleClicked.connect(self._show_selected_scene)

    def eventFilter(self, source, event):
        '''
        Event filter for capturing and processing events.
        '''
        focus_event = 24  # This will be replaced with the event itself.
        paint_event = 12
        if event.type() == paint_event:
            self._check_table_state()
        elif event.type() == focus_event:
            self._populate_data()

        return super().eventFilter(source, event)

    def _populate_data(self):
        '''
        Main loop for populating the table and class data structures.
        '''
        # Clear rows
        self.sceneTableWidget.setRowCount(0)
        self.scene_list = self._update_scene_list()
        self._update_scene_dict()

        self._format_condensed_widget()

        for scene_name in self.scene_list:
            # Add a row for the scene
            row_number = self.sceneTableWidget.rowCount()
            self.sceneTableWidget.insertRow(row_number)
            self.sceneTableWidget.setRowHeight(row_number, 100)
            # Get data from scene_dict
            scene_image_label = self._get_scene_image_label(
                self.scene_dict[scene_name][SceneDictIndex.QPIXMAP])
            scene_message = self.scene_dict[scene_name][SceneDictIndex.MESSAGE]
            scene_actions = self.scene_dict[scene_name][SceneDictIndex.ACTIONS]

            # Set items within the new row
            self.sceneTableWidget.setItem(
                row_number, SceneTableColumn.NAME,
                QtWidgets.QTableWidgetItem(scene_name))
            self.sceneTableWidget.setCellWidget(
                row_number, SceneTableColumn.IMAGE, scene_image_label)
            self.sceneTableWidget.setItem(
                row_number, SceneTableColumn.MESSAGE,
                QtWidgets.QTableWidgetItem(scene_message))
            self.sceneTableWidget.setItem(
                row_number, SceneTableColumn.ACTIONS,
                QtWidgets.QTableWidgetItem(scene_actions))

        self._format_table()

        self.sceneTableWidget.selectionModel().clearSelection()

    def _update_scene_dict(self):
        '''
        Updates the class' scene_dict to reflect the current state of
        the class' scene_list. Currently, place holder value are being used
        for the message and action but these will be replaced by functions
        responsible for fetching these values for a given scene.

        Note: This is also currently being used to make sure that scenes
        are being properly removed from the dictionary if they no longer
        appear in the scene_list. There may be a better way of doing this or
        this may be handled in separate step in the future.
        '''
        for scene_position in range(len(self.scene_list)):
            scene_name = self.scene_list[scene_position]
            if scene_name not in self.scene_dict:
                scene_pix_map = self._get_scene_png(scene_name)
                scene_message = 'This is a base message'
                scene_actions = 'Rock, zoom, something'
                self.scene_dict[scene_name] = [
                    scene_pix_map, scene_message,
                    scene_actions, scene_position]

        remove_scenes = [
            scene_name for scene_name in self.scene_dict
            if scene_name not in self.scene_list]

        for scene_name in remove_scenes:
            self.scene_dict.pop(scene_name)

    def _get_scene_png(self, scene_name):
        '''
        Returns a QPixmap object that can be assigned to a label
        in order to display the scene thumbnail.
        This will be obsolete and removed in the future.
        '''
        scene_pix_map = QtGui.QPixmap()
        png_buf = self.cmd.get_scene_thumbnail(scene_name)
        scene_pix_map.loadFromData(png_buf, "PNG")
        return scene_pix_map

    def _get_scene_image_label(self, scene_pix_map):
        '''
        Takes in the QPixmap from get_scene_png and uses it to create a label
        displaying the thumbnail.
        In the future, this function will generate its
        own QPixmap object from the PNG buffer stored with the scene.
        '''
        scene_image_label = QtWidgets.QLabel(self)
        scene_image_label.setPixmap(scene_pix_map)
        return scene_image_label

    def _format_condensed_widget(self):
        '''
        Formats the widget to look and size properly. There will be a
        matching function for the expanded view of the widget.
        '''
        self.sceneTableWidget.setMinimumWidth(275)
        self.setMinimumWidth(375)
        self.resize(300, self.height())
        self.sceneTableWidget.setColumnCount(2)
        self.sceneTableWidget.setHorizontalHeaderLabels(
            ['Name', 'Scene Preview'])
        self.sceneTableWidget.resize(260, 500)

    def _format_table(self):
        '''
        Formats the table to look properly. Currently, vertical headers
        are being hidden but they will be used for reordering.
        '''
        hh = self.sceneTableWidget.horizontalHeader()
        hh.setStretchLastSection(True)
        self.sceneTableWidget.setFocus()
        self.sceneTableWidget.hide()
        self.sceneTableWidget.resizeColumnsToContents()
        self._set_vertical_headers()
        self.sceneTableWidget.show()

    def _set_vertical_headers(self):
        # Get row count
        row_count = self.sceneTableWidget.rowCount()

        # Iterate in range, setting item
        for ind in range(row_count):
            # Create  arrows item
            arrows = QtWidgets.QTableWidgetItem('\u2195')
            arrows_font = arrows.font()
            arrows_font.setPointSize(20)
            arrows.setFont(arrows_font)
            self.sceneTableWidget.setVerticalHeaderItem(ind, arrows)

    def _update_scene_list(self):
        '''
        Returns an updated list of the current scenes. Currently, this
        is very basic but some handling will be done here in the future
        or this will likely be removed.
        '''
        list_scenes = self.cmd.get_scene_list()
        return list_scenes

    def _add_scene(self):
        '''
        Calls cmd.scene with append and repopulates the table.
        Table is scrolled to the bottom to show new scene.
        '''
        self.cmd.scene('new', 'append', quiet=0)
        self._populate_data()
        self.sceneTableWidget.scrollToBottom()

    def _update_scene(self):
        '''
        Calls cmd.scene with update, replacing the selected scene with
        new scene. Image is not currently being updated but it should.
        Button only active with selection.
        '''
        selection = self.sceneTableWidget.selectionModel().selectedIndexes()
        if selection:
            name = selection[0].data()
            self.cmd.scene(name, 'update')
            self.scene_dict[name][SceneDictIndex.QPIXMAP] = self._get_scene_png(
                name)
            self._update_table()

    def _delete_scene(self):
        '''
        Calls cmd.scene with clear, removing the selected scene.
        Button only active with selection.
        '''
        selection = self.sceneTableWidget.selectionModel().selectedIndexes()
        name = selection[0].data()
        self.cmd.scene(name, 'clear')
        try:
            self.scene_dict.pop(name)
            self.scene_list.remove(name)
        except Exception as e:
            print("Item not found")
            print(e)
        self._update_table()

    def _show_selected_scene(self):
        '''
        Calls cmd.scene with recall, recalling scene to workspace.
        '''
        selection = self.sceneTableWidget.selectionModel().selectedIndexes()
        if selection:
            name = selection[0].data()
            self.cmd.scene(name, 'recall')

    def _update_table(self):
        '''
        Currently pointless but scrolling/formatting will happen
        here to make updating the table more seemless after
        intially populating.
        '''
        self._populate_data()

    def _compare_scene_lists(self, old_list, new_list):
        '''
        Compares current and new list of scenes.
        Returns list of elements that differ ([old, new])
        '''
        diff_list = []
        for i in range(min(len(old_list), len(new_list))):
            if new_list[i] != old_list[i]:
                diff_list.append([old_list[i], new_list[i]])
        return diff_list

    def _get_table_scene_list(self):
        '''
        Returns a sorted list of scene names based on the current
        state of the table. This will be a critical part of reordering,
        but also used for detecting renaming.
        '''
        scene_coor_list = []
        for row_num in range(self.sceneTableWidget.rowCount()):
            name = self.sceneTableWidget.model().data(
                self.sceneTableWidget.model().index(row_num, 0))
            position = int(
                self.sceneTableWidget.rowViewportPosition(row_num)/100)

            scene_coor_list.append(SceneWithTablePosition(name, position))

        scene_coor_list.sort(key=lambda scene: scene.position)

        return [scene_with_pos.name for scene_with_pos in scene_coor_list]

    def _check_table_state(self):
        '''
        Checks the current state of the table for reordering.
        '''
        scene_ordered_list = self._get_table_scene_list()
        self.scene_list = self.cmd.get_scene_list()
        diff_list = self._compare_scene_lists(
            self.scene_list, scene_ordered_list)
        if len(diff_list) > 1:
            # This is the reordering case since multiple have changed
            self._reorder_scenes(scene_ordered_list)

    def _selection_changed(self):
        '''
        Enables buttons when selection is made.
        '''
        item_selected = bool(
            self.sceneTableWidget.selectionModel().selectedIndexes())
        self.deleteButton.setEnabled(item_selected)
        self.updateButton.setEnabled(item_selected)

    def _rename_scene(self, item):
        '''
        Takes in an item from _item_changed and checks conditions before
        calling scene with 'rename'.
        '''
        if ' ' in item.text():
            print("Scene names with spaces are not supported")
            self._update_table()
        elif not item.text():
            print("Blank scene names are not allowed")
            self._update_table()
        else:
            diff_list = self._compare_scene_lists(
                self.cmd.get_scene_list(),
                self._get_table_scene_list())
            if diff_list:
                for old_scene, new_scene in diff_list:
                    self.cmd.scene(old_scene, 'rename', new_key=new_scene)

    def _reorder_scenes(self, scene_ordered_list):
        '''
        Calls cmd.scene_order which requires a string of scene names.
        This string is created using the ordered scene list generated
        by _check_table_state.
        '''
        scene_order_string = ' '.join(
            scene for scene in scene_ordered_list if scene in self.scene_list)
        self.cmd.scene_order(scene_order_string)
        self.scene_list = self._update_scene_list()

    def _item_changed(self, item):
        """
        Called every time an item in the table is changed.
        @param item: The item which has changed
        @type  item: QStandardItem
        """
        if item.column() == SceneTableColumn.NAME:
            self._rename_scene(item)



#A* -------------------------------------------------------------------
#B* This file contains source code for the PyMOL computer program
#C* Copyright (c) Schrodinger, LLC
#D* -------------------------------------------------------------------
#E* It is unlawful to modify or remove this copyright notice.
#F* -------------------------------------------------------------------
#G* Please see the accompanying LICENSE file for further information.
#H* -------------------------------------------------------------------
#I* Additional authors of this source file include:
#-*
#-*
#-*
#Z* -------------------------------------------------------------------

import sys
cmd_module = __import__("sys").modules["pymol.cmd"]

import pymol

def clean(selection, present='', state=-1, fix='', restrain='',
          method='mmff', async_=0, save_undo=1, message=None,
          _self=cmd_module, **kwargs):
    '''
DESCRIPTION

    Run energy minimization on the given selection, using an MMFF94
    force field.
    '''
    raise pymol.IncentiveOnlyException



# A* -------------------------------------------------------------------
# B* This file contains source code for the PyMOL computer program
# C* Copyright (c) Schrodinger, LLC.
# D* -------------------------------------------------------------------
# E* It is unlawful to modify or remove this copyright notice.
# F* -------------------------------------------------------------------
# G* Please see the accompanying LICENSE file for further information.
# H* -------------------------------------------------------------------
# I* Additional authors of this source file include:
# -*
# -*
# -*
# Z* -------------------------------------------------------------------

import dataclasses
import math
from typing import List, Dict

from chempy import cpv

from pymol import cgo

from . import cmd


@dataclasses.dataclass
class Point:
    """
    Represents a 3D point in space

    Attributes:
        x (float): x-coordinate
        y (float): y-coordinate
        z (float): z-coordinate
    """
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def array(self) -> List[float]:
        """
        Converts point into 3-component array
        :returns : array of xyz coordinates
        """
        return [self.x, self.y, self.z]


@dataclasses.dataclass
class Color:
    """
    Represents a Color in RGBA space

    Attributes:
        r (float): red channel value
        g (float): green channel value
        b (float): blue channel value
        a (float): alpha channel value
    """
    r: float = 0.0
    g: float = 0.0
    b: float = 0.0
    a: float = 1.0

    def array(self) -> List[float]:
        """
        Converts color into 3-component array
        """
        return [self.r, self.g, self.b]


@dataclasses.dataclass
class CGO:
    """
    Base class for all CGOs

    For all CGO types, whenever attributes are changed, `rebuild` should be
    called to update the data cache.
    """
    _data: List[float] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        """
        Rebuilds CGO. All child classes should implement this.
        """
        self.rebuild()

    @property
    def data(self):
        return self._data


@dataclasses.dataclass
class Sphere(CGO):
    """
    Sphere CGO

    Attributes:
        center (Point): center position
        radius (float): sphere radius
        color (Color): sphere color
    """
    center: Point = Point(0, 0, 0)
    radius: float = 0.0
    color: Color = Point(0, 0, 0)

    def rebuild(self) -> None:
        """
        Rebuilds sphere
        """
        self._data.clear()
        self._data.append(cgo.COLOR)
        self._data.extend(self.color.array())
        self._data.append(cgo.SPHERE)
        self._data.extend(self.center.array())
        self._data.append(self.radius)


@dataclasses.dataclass
class Cylinder(CGO):
    """
    Cylinder CGO

    Attributes:
        point1 (Point): center position of first side
        point2 (Point): center position of second side
        radius (float): sphere radius
        color1 (Color): color of first cylinder half
        color2 (Color): color of second cylinder half
    """
    point1: Point = Point(0, 0, 0)
    point2: Point = Point(1, 1, 1)
    radius: float = 1.0
    color1: Color = Point(0, 0, 0)
    color2: Color = Color(0, 0, 0)

    def rebuild(self) -> None:
        """
        Rebuilds cylinder
        """
        self._data.clear()
        self._data.append(cgo.CYLINDER)
        self._data.extend(self.point1.array())
        self._data.extend(self.point2.array())
        self._data.append(self.radius)
        self._data.extend(self.color1.array())
        self._data.extend(self.color2.array())


@dataclasses.dataclass
class Torus(CGO):
    """
    Torus CGO

    Attributes:
        center (Point): torus center position
        normal (Point): torus normal direction
        radius (float): ring radius
        color (Color): torus color
        cradius (float): torus ring radius
        samples (int): number of samples to generate torus segment
        csamples (int): number of samples to generate torus ring
    """
    center: Point = Point(0.0, 0.0, 0.0)
    normal: Point = Point(0.0, 0.0, 1.0)
    radius: float = 1.0
    color: Color = Color(0.0, 1.0, 1.0, 1.0)
    cradius: float = 0.25
    samples: int = 20
    csamples: int = 20

    def rebuild(self) -> None:
        """
        Rebuilds torus
        """
        obj = []

        axis = cpv.cross_product(self.normal.array(), (0., 0., 1.))
        angle = -cpv.get_angle(self.normal.array(), (0., 0., 1.))
        matrix = cpv.rotation_matrix(angle, cpv.normalize(axis))

        def obj_vertex(x, y, z):
            return [cgo.VERTEX] + cpv.add(self.center.array(),
                                          cpv.transform(matrix, [x, y, z]))

        def obj_normal(x, y, z):
            return [cgo.NORMAL] + cpv.transform(matrix, [x, y, z])

        r = self.radius
        cr = self.cradius
        rr = 1.5 * cr
        dv = 2 * math.pi / self.csamples
        dw = 2 * math.pi / self.samples
        v = 0.0
        w = 0.0

        while w < 2 * math.pi:
            v = 0.0
            c_w = math.cos(w)
            s_w = math.sin(w)
            c_wdw = math.cos(w + dw)
            s_wdw = math.sin(w + dw)

            obj.append(cgo.BEGIN)
            obj.append(cgo.TRIANGLE_STRIP)

            obj.append(cgo.COLOR)
            obj.extend(self.color.array())

            while v < 2 * math.pi + dv:
                c_v = math.cos(v)
                s_v = math.sin(v)
                c_vdv = math.cos(v + dv)
                s_vdv = math.sin(v + dv)
                obj.extend(
                    obj_normal((r + rr * c_v) * c_w - (r + cr * c_v) * c_w,
                               (r + rr * c_v) * s_w - (r + cr * c_v) * s_w,
                               (rr * s_v - cr * s_v)))
                obj.extend(
                    obj_vertex((r + cr * c_v) * c_w, (r + cr * c_v) * s_w,
                               cr * s_v))
                obj.extend(
                    obj_normal(
                        (r + rr * c_vdv) * c_wdw - (r + cr * c_vdv) * c_wdw,
                        (r + rr * c_vdv) * s_wdw - (r + cr * c_vdv) * s_wdw,
                        rr * s_vdv - cr * s_vdv))
                obj.extend(
                    obj_vertex((r + cr * c_vdv) * c_wdw,
                               (r + cr * c_vdv) * s_wdw, cr * s_vdv))
                v += dv

            obj.append(cgo.END)
            w += dw

        self._data = obj


@dataclasses.dataclass
class Cone(CGO):
    """
    Cone CGO

    Attributes:
        tip (Point): cone tip position
        base_center (Point): center position of cone base
        radius (float): cone base radius
        color (Color): cone color
    """
    tip: Point = Point(1.0, 0.0, 0.0)
    base_center: Point = Point(0.0, 0.0, 0.0)
    radius: float = 1.0
    color: Color = Color(0.0, 0.0, 0.0)

    def rebuild(self) -> None:
        """
        Rebuilds cone
        """
        self._data = []
        self._data.append(cgo.CONE)
        self._data.extend(self.tip.array())
        self._data.extend(self.base_center.array())
        self._data.extend([self.radius, 0.0])
        self._data.extend(self.color.array())
        self._data.extend(self.color.array())
        self._data.extend([1.0, 0.0])


@dataclasses.dataclass
class Bezier(CGO):
    """
    Cubic Bezier CGO

    Attributes:
        control_pt_A (Point): on-curve bezier spline position (from)
        A_right_handle (Point): influences curve from A to B
        control_pt_B (Point): on-curve bezier spline position (to)
        B_left_handle (Point): influences curve from A to B
        A_color (Color): color at point A
        B_color (Color): color at point B
    """
    control_pt_A = Point(-5.0, 0.0, 0.0)
    A_right_handle = Point(0.0, 10.0, 0.0)
    B_left_handle = Point(1.0, -10.0, 0.0)
    control_pt_B = Point(5.0, 0.0, 0.0)

    def rebuild(self) -> None:
        """
        Rebuilds bezier spline
        """
        self._data = []
        self._data.append(cgo.BEZIER)
        self._data.extend(self.control_pt_A.array())
        self._data.extend(self.A_right_handle.array())
        self._data.extend(self.B_left_handle.array())
        self._data.extend(self.control_pt_B.array())


class CGOBuilder:
    """
    PyMOL CGO Builder

    Holds a storage of CGOs to load into PyMOL as a single object.

    Attributes:
        cgos (Dict[CGO]): Collection of CGOs
    """
    cgos: Dict[int, CGO] = {}
    _token_count = 0

    def add(self, cgo: CGO) -> int:
        """
        Adds CGO to builder

        :param cgo (CGO): cgo to add

        :return: token ID for added CGO
        """
        token = self._token_count
        self.cgos[token] = cgo
        self._token_count += 1
        return token

    def remove(self, token: int) -> None:
        """
        Removes CGO from builder

        :param token (int): token of CGO to remove
        """
        self.cgos.pop(token, None)

    def clear(self) -> None:
        """
        Clears the CGO Collection
        """
        self.cgos.clear()

    def load(self, name: str, flush: bool = True, _self=cmd) -> None:
        """
        Loads the CGO as an object into pymol

        :param name (str): Name of object
        :param flush (bool): Should reset the object data cache after loading
        """
        all_serialized = []
        for _, cgo_obj in self.cgos.items():
            all_serialized.extend(cgo_obj.data)

        _self.load_cgo(all_serialized, name)
        if flush:
            self.cgos.clear()



import json
import os
from textwrap import fill

from pymol import setting
from pymol import save_shortcut
from pymol.Qt import QtGui, QtWidgets
from pymol.Qt import QtCore, QtCoreModels
from pymol.shortcut_manager import ShortcutManager, ShortcutIndex
from pymol.keyboard import get_default_keys
Qt = QtCore.Qt
QSI = QtGui.QStandardItem  # For brevity


def get_shortcut_key_map():
    shortcut_key_map = {}
    for key, value in vars(Qt).items():
        if isinstance(value, Qt.Key):
            shortcut_key_map[value] = key.partition('_')[2]
    return shortcut_key_map


_SHORTCUT_KEY_MAP = get_shortcut_key_map()

_SHORTCUT_MODIFIER_MAP = {
    Qt.ControlModifier: _SHORTCUT_KEY_MAP[Qt.Key_Control],
    Qt.AltModifier: _SHORTCUT_KEY_MAP[Qt.Key_Alt],
    Qt.ShiftModifier: _SHORTCUT_KEY_MAP[Qt.Key_Shift],
    Qt.MetaModifier: _SHORTCUT_KEY_MAP[Qt.Key_Meta],
}

_REPLACE_KEYS = {
    'PageUp': 'pgup',
    'PageDown': 'pgdn',
    'Home': 'home',
    'Insert': 'insert',
    'Up': 'up',
    'Down': 'down',
    'Left': 'left',
    'Right': 'right',
    'End':'end'
}

class PyMOLShortcutMenu(QtWidgets.QWidget):
    '''
    Keyboard shortcut dialog for PyMOL. This displays all assigned shortcuts
    and allows them to be changed or new shortcuts to be created.
    '''

    def __init__(self, parent, saved_shortcuts, cmd):
        QtWidgets.QWidget.__init__(self, parent, Qt.Window)
        self.resize(700, 700)
        self.cmd = cmd
        self.shortcut_manager = ShortcutManager(saved_shortcuts, cmd)

        self.build_panel_elements(parent)

    def build_panel_elements(self, parent):
        '''
        Responsible for creating all panel elements in order and adding them to the layout.
        '''
        self.create_new_form = parent.load_form("create_shortcut", None)
        self.help_form = parent.load_form("help_shortcut", None)
        self.confirm_change = parent.load_form("change_confirm", None)

        self.model = QtGui.QStandardItemModel(self)
        self.proxy_model = QtCoreModels.QSortFilterProxyModel(self)
        self.proxy_model.setSourceModel(self.model)
        self.proxy_model.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.proxy_model.setFilterKeyColumn(-1)

        self.setWindowTitle('Keyboard Shortcut Menu')
        layout = QtWidgets.QVBoxLayout(self)
        self.setLayout(layout)

        # Create layout for filter bar and refresh button
        top_layout = QtWidgets.QGridLayout()
        layout.addLayout(top_layout)

        # Filter
        self.filter_le = QtWidgets.QLineEdit(self)
        top_layout.addWidget(self.filter_le)
        self.filter_le.setPlaceholderText("Filter")
        self.filter_le.textChanged.connect(self.proxy_model.setFilterRegExp)

        self.refresh_button = QtWidgets.QPushButton(self)
        self.refresh_button.resize(26, 26)
        top_layout.addWidget(self.refresh_button, 0, 1)
        # themed icons only available by default on X11
        if self.refresh_button.icon().isNull():
            self.refresh_button.setIcon(QtGui.QIcon(
                os.path.expandvars('$PYMOL_DATA/pmg_qt/icons/refresh.svg')))
        self.refresh_button.setToolTip(
            "Refresh the table to reflect any external changes")
        self.refresh_button.clicked.connect(self.refresh_populate)

        # Table
        self.table = QtWidgets.QTableView(self)
        self.table.setModel(self.proxy_model)
        layout.addWidget(self.table)
        self.intial_populate()
        self.formatTable()

        # Add layout for buttons
        button_layout = QtWidgets.QGridLayout()
        layout.addLayout(button_layout)

        # Buttons
        self.create_new_button = QtWidgets.QPushButton(self)
        button_layout.addWidget(self.create_new_button, 0, 0)
        self.create_new_button.setText("Create New")
        self.create_new_button.setToolTip(
            "Add a key binding that does not currently appear on the table")
        self.create_new_button.clicked.connect(
            lambda: self.create_new_form._dialog.show())

        self.delete_selected_button = QtWidgets.QPushButton(self)
        button_layout.addWidget(self.delete_selected_button, 0, 1)
        self.delete_selected_button.setText("Delete Selected")
        self.delete_selected_button.setToolTip(
            "Unbind selected key bindings and remove any that have been created")
        self.delete_selected_button.clicked.connect(self.delete_selected)
        self.delete_selected_button.setEnabled(False)

        self.reset_selected_button = QtWidgets.QPushButton(self)
        button_layout.addWidget(self.reset_selected_button, 0, 2)
        self.reset_selected_button.setText("Reset Selected")
        self.reset_selected_button.setToolTip(
            "Restore selected key bindings to their default values")
        self.reset_selected_button.clicked.connect(self.reset_selected)
        self.reset_selected_button.setEnabled(False)

        self.reset_all_button = QtWidgets.QPushButton(self)
        button_layout.addWidget(self.reset_all_button, 0, 3)
        self.reset_all_button.setText("Reset All")
        self.reset_all_button.setToolTip(
            "Restore all key bindings to their default values and remove any that have been created")
        self.reset_all_button.clicked.connect(self.reset_all_default)

        self.save_button = QtWidgets.QPushButton(self)
        button_layout.addWidget(self.save_button, 0, 4)
        self.save_button.setText("Save")
        self.save_button.setToolTip(
            "Save the current key bindings to be loaded automatically when opening PyMOL")
        self.save_button.clicked.connect(self.shortcut_manager.save_shortcuts)

        # Ensuring that confirmed key and binding remain in scope
        self.confirm_new_key = ''
        self.confirm_new_binding = ''

        # Connect create new and confirm menus
        self.create_new_shortcut_menu_connect()
        self.confirm_menu_connect()

        self.model.itemChanged.connect(self.itemChanged)

    def populateData(self):
        '''
        Fill the model with data from shortcut_dict.
        '''
        self.model.clear()
        self.model.setHorizontalHeaderLabels(
            ['Key', 'Command (click to edit)', 'Description'])

        for key, shortcut_list in self.shortcut_manager.cmd.shortcut_dict.items():
            key_item = QSI(key)
            command_item = QSI()
            descript_item = QSI()
            key_item.setFlags(Qt.ItemIsEnabled)
            descript_item.setFlags(Qt.ItemIsEditable)

            if shortcut_list[ShortcutIndex.USER_DEF]:
                if shortcut_list[ShortcutIndex.USER_DEF] != "Deleted":
                    command_text = shortcut_list[ShortcutIndex.USER_DEF]
                    descript_text = "user defined"
                else:
                    command_text = "Deleted"
                    descript_text = "Deleted"
            else:
                command_text = shortcut_list[ShortcutIndex.COMMAND]
                descript_text = shortcut_list[ShortcutIndex.DESCRIPT]

            command_item.setText(command_text)
            descript_item.setText(descript_text)

            self.model.appendRow([key_item, command_item, descript_item])
        self.formatTable()
        self.table.selectionModel().selectionChanged.connect(self.selection_changed)

    def selection_changed(self):
        item_selected = bool(self.table.selectionModel().selectedIndexes())
        self.delete_selected_button.setEnabled(item_selected)
        self.reset_selected_button.setEnabled(item_selected)

    def intial_populate(self):
        '''
        Runs when the menu is first opened. Separate from populateData so that
        the saved dictionary and current state of key_mappings can be checked.
        '''
        self.shortcut_manager.check_saved_dict()
        self.shortcut_manager.check_key_mappings()
        self.populateData()

    def refresh_populate(self):
        '''
        Called through the refresh button. 
        '''
        self.shortcut_manager.check_key_mappings()
        self.populateData()

    def reset_all_default(self):
        '''
        Iterates over all values to restore their default commands.
        This will restore them to the values from keyboard.py
        '''
        self.shortcut_manager.reset_all_default()
        self.populateData()

    def delete_selected(self):
        '''
        Removes selected keybindings and updates table to say "Deleted".
        Keys that don't have default values will be removed completely.
        '''
        selection_model = self.table.selectionModel()
        list_indexes = selection_model.selectedIndexes()
        delete_keys = []

        for ind, table_index in enumerate(list_indexes):
            table_colm_key = self.table.model().index(table_index.row(), 0)
            table_colm_command = self.table.model().index(table_index.row(), 1)
            table_colm_descipt = self.table.model().index(table_index.row(), 2)

            delete_key = table_colm_key.data()
            self.cmd.set_key(delete_key, '')
            if delete_key in self.shortcut_manager.default_bindings:
                self.table.model().setData(table_colm_command, 'Deleted')
                self.table.model().setData(table_colm_descipt, 'Deleted')
                self.shortcut_manager.cmd.shortcut_dict[delete_key][ShortcutIndex.USER_DEF] = 'Deleted'
            else:
                self.model.removeRow(table_index.row())
                print(delete_key, " has been deleted and will be removed from the table")
                delete_keys.append(delete_key)

        for key in delete_keys:
            del self.shortcut_manager.cmd.shortcut_dict[key]

    def reset_selected(self):
        '''
        Restores default key bindings for items selected in the tables selection model.
        '''
        selection_model = self.table.selectionModel()
        list_indexes = selection_model.selectedIndexes()

        for ind, table_index in enumerate(list_indexes):
            table_colm_key = self.table.model().index(table_index.row(), 0)
            table_colm_command = self.table.model().index(table_index.row(), 1)
            table_colm_descipt = self.table.model().index(table_index.row(), 2)

            reset_key = table_colm_key.data()
            if reset_key not in self.shortcut_manager.default_bindings:
                print("This key does not have a default value.")
            else:
                reset_binding = self.shortcut_manager.default_bindings[reset_key]
                reset_command = self.shortcut_manager.cmd.shortcut_dict[
                    reset_key][ShortcutIndex.COMMAND]
                reset_description = self.shortcut_manager.cmd.shortcut_dict[
                    reset_key][ShortcutIndex.DESCRIPT]

                self.table.model().setData(table_colm_command, reset_command)
                self.table.model().setData(table_colm_descipt, reset_description)

                self.shortcut_manager.cmd.shortcut_dict[reset_key][ShortcutIndex.USER_DEF] = ''

                self.cmd.set_key(reset_key, reset_binding)

    def create_new_shortcut_menu_connect(self):
        self.create_new_form.createButton.clicked.connect(
            self.create_new_shortcut_caller)
        self.create_new_form.helpButton.clicked.connect(
            self.help_menu_shortcut)
        self.create_new_form.keyEdit.installEventFilter(self)
        self.create_new_form.helpButton.setDefault(False)
        self.create_new_form.helpButton.setAutoDefault(False)

    def eventFilter(self, source, event):
        '''
        Event filter for creating new shortcuts. Processes the key event before passing it on.
        '''
        if (event.type() == QtCore.QEvent.KeyPress and source is self.create_new_form.keyEdit):
            raw_string = self.keyevent_to_string(event)
            processed_string = self.process_keyevent_string(raw_string)

            if processed_string in self.shortcut_manager.reserved_keys:
                return 0

            if processed_string:
                self.create_new_form.keyEdit.setText(processed_string)

        return super().eventFilter(source, event)

    def keyevent_to_string(self, event):
        '''
        Generates string from captured key event for process_keyevent_string.
        '''
        keyevent_list = []
        for mod, event_text in _SHORTCUT_MODIFIER_MAP.items():
            if event.modifiers() & mod:
                keyevent_list.append(event_text)

        key = _SHORTCUT_KEY_MAP.get(event.key(), event.text())

        if key not in keyevent_list:
            keyevent_list.append(key)

        return ' '.join(keyevent_list)

    def process_keyevent_string(self, raw_string):
        '''
        Returns string of keyevent used to populate create key menu.
        Provides filtering to the keyevents, returning a list of the accepted strings.
        raw_string: string from keyevent_to_string 
        '''
        split_string = raw_string.split()
        process_list = []
        prefix_key = split_string[0]
        if len(split_string) >= 2:
            prefix_key = split_string[0]
            suffix_key = split_string[1]
            if (prefix_key == 'Control' or prefix_key == 'Meta') and split_string[1]:
                if suffix_key == 'Shift' and len(split_string) > 2:
                    process_list.append('CTSH')
                    suffix_key = split_string[2]
                else:
                    process_list.append('CTRL')
            elif prefix_key == 'Alt':
                process_list.append('ALT')
            elif prefix_key == 'Shift':
                process_list.append('SHFT')
            if suffix_key in _REPLACE_KEYS:
                suffix_key = _REPLACE_KEYS[suffix_key]
            process_list.append(suffix_key)
        elif prefix_key in _REPLACE_KEYS:
            process_list.append(_REPLACE_KEYS[prefix_key])
        return('-'.join(process_list))

    def help_menu_shortcut(self):
        self.help_form._dialog.show()

    def confirm_menu_connect(self):
        self.confirm_change.confirmButton.clicked.connect(
            lambda: self.shortcut_manager.create_new_shortcut(self.confirm_new_key, self.confirm_new_binding))
        self.confirm_change.confirmButton.clicked.connect(lambda: self.populateData())
        self.confirm_change.confirmButton.clicked.connect(
            lambda: self.confirm_change._dialog.hide())
        self.confirm_change.cancelButton.clicked.connect(
            lambda: self.confirm_change._dialog.hide())

    def create_new_shortcut_caller(self):
        '''
        Creates a new shortcut after checking existing and reserved keys.
        '''
        new_key = self.create_new_form.keyEdit.text()
        new_binding = self.create_new_form.commandEdit.text()

        if new_key == '':
            pass
        elif new_key in self.shortcut_manager.cmd.shortcut_dict:
            hide_confirm_menu = self.confirm_change.doNotShowCheckBox.isChecked()
            if not hide_confirm_menu:
                self.confirm_new_key = new_key
                self.confirm_new_binding = new_binding
                self.confirm_change._dialog.show()
            else:
                self.shortcut_manager.create_new_shortcut(new_key, new_binding)
                self.populateData()
        else:
            self.shortcut_manager.create_new_shortcut(new_key, new_binding)
            self.populateData()
            self.table.scrollToBottom()

    def formatTable(self):
        '''
        Set up the table to look appropriately
        '''
        hh = self.table.horizontalHeader()
        hh.setStretchLastSection(True)
        self.table.verticalHeader().setVisible(False)
        self.table.setFocus()
        self.table.hide()
        self.table.resizeColumnsToContents()
        self.table.show()

    def itemChanged(self, item):
        """
        Called every time an item in the table is changed, only command items
        are changeable.
        @param item: The item which has changed
        @type  item: QStandardItem
        """
        try:
            if item.column() == 1 and item.text() != "Deleted":
                changed_key = self.model.index(item.row(), 0).data()
                changed_index = self.table.model().index(item.row(), 0)

                self.cmd.set_key(changed_key, item.text())
                self.shortcut_manager.cmd.shortcut_dict[changed_key][2] = item.text()

                filter_active = bool(self.filter_le.text())
                if not filter_active:
                    self.table.model().setData(self.table.model().index(item.row(), 2), 'user defined')
                else:
                    pass
        except Exception as e:
            print(e)
            print("Failed to change key binding")



'''
Volume color ramp utilities
'''

import sys
from pymol.constants import CURRENT_STATE

cmd = sys.modules["pymol.cmd"]

_volume_windows = {}
_volume_windows_qt = {}


def peak(v, c, a=0.2, d=0.2):
    return [v - d, c, 0., v, c, a, v + d, c, 0.]

namedramps = {
    '2fofc': [
         1.,  "blue",   .0,
         1.,  "blue",   .2,
         1.4, "blue",   .0,
    ],
    'fofc': [
        -3.5, "red",   .0,
        -3.,  "red",   .2,
        -2.9, "red",   .0,
         2.9, "green", .0,
         3.,  "green", .2,
         3.5, "green", .0,
    ],
    'esp': [
        -1.2, "red",  .0,
        -1.,  "red",  .2,
        -1.,  "0xff9999", .0,
         1.,  "0x9999ff", .0,
         1.,  "blue", .2,
         1.2, "blue", .0,
    ],
    'rainbow': [
        1.0, 'blue', 0.1,
        1.5, 'cyan', 0.1,
        2.0, 'green', 0.1,
        2.5, 'yellow', 0.1,
        3.0, 'orange', 0.1,
        3.5, 'red', 0.1,
    ],
    'rainbow2':
        peak(1.0, 'blue') +
        peak(1.5, 'cyan') +
        peak(2.0, 'green') +
        peak(2.5, 'yellow') +
        peak(3.0, 'orange') +
        peak(3.5, 'red'),
}

def volume_ramp_new(name, ramp):
    '''
DESCRIPTION

    Register a named volume ramp which can be used as a preset
    when creating or coloring volumes. The name will appear in the
    internal menu at "A > volume" and "C".

ARGUMENTS

    name = string: name of the new ramp

    ramp = list: space delimited list of value, color, alpha

EXAMPLE

    volume_ramp_new pink1sigma, \\
       0.9 violet 0.0 \\
       1.0 magenta 0.3 \\
       1.5 pink 0.0

SEE ALSO

    volume, volume_color
    '''
    from .checking import is_string
    if is_string(ramp):
        ramp = ramp.split()
    namedramps[name] = ramp


def get_volume_color(name, state=CURRENT_STATE, quiet=1, _self=cmd):
    '''
DESCRIPTION

    Get the volume color ramp of a volume object.

ARGUMENTS

    name = str: volume object name

    state = int: state of volume
    '''
    quiet = int(quiet)

    r = _self.DEFAULT_ERROR
    with _self.lockcm:
        r = _self._cmd.get_volume_ramp(_self._COb, name, int(state)-1)

    if isinstance(r, list):
        if not quiet:
            import random
            rname = 'ramp%03d' % random.randint(0, 999)
            print('### cut below here and paste into script ###')
            print('cmd.volume_ramp_new(%s, [\\' % repr(rname))
            for i in range(0, len(r), 5):
                print('    %6.2f, %.2f, %.2f, %.2f, %.2f, \\' % tuple(r[i:i+5]))
            print('    ])')
            print('### cut above here and paste into script ###')
        elif quiet < 0:
            print('volume_color %s, ' % (name), end=' ')
            for i in range(0, len(r), 5):
                print('\\\n    %6.2f %.2f %.2f %.2f %.2f' % tuple(r[i:i+5]), end=' ')
            print('')

    return r

def volume_color(name, ramp='', state=CURRENT_STATE, quiet=1, _guiupdate=True, _self=cmd):
    '''
DESCRIPTION

    Set or get the volume colors.

ARGUMENTS

    name = str: volume object name

    ramp = str, list or empty: named ramp, space delimited string or list
    with (x, color, alpha, ...) or (x, r, g, b, alpha, ...) values. If empty, get
    the current volume colors.

    state = int: state of volume to color

EXAMPLE

    fetch 1a00, map, type=2fofc
    volume vol, map
    volume_color vol, .8 cyan 0. 1. blue .3 2. yellow .3
    '''
    quiet = int(quiet)

    if not ramp:
        return get_volume_color(name, state, quiet, _self)

    if isinstance(ramp, str) and ramp in namedramps:
        ramp = namedramps[ramp]

    ramplist = ramp_expand(ramp)

    with _self.lockcm:
        r = cmd._cmd.set_volume_ramp(_self._COb, name, ramplist, int(state)-1)

    if _guiupdate and name in _volume_windows:
        from pymol import gui
        def func():
            import tkinter as Tkinter
            try:
                panel = _volume_windows[name].panel
                panel.set_flat(ramplist)
            except (LookupError, Tkinter.TclError):
                pass
        app = gui.get_pmgapp()
        app.execute(func)

    if _guiupdate and name in _volume_windows_qt:
        from pymol import gui
        app = gui.get_pmgapp()
        @app.execute
        def _():
            try:
                panel = _volume_windows_qt[name]
                panel.widget().editor.setColors(ramplist)
            except LookupError:
                pass

    return r

def volume_panel(name, quiet=1, _self=cmd, _noqt=0):
    '''
DESCRIPTION

    Open an interactive volume ramp panel

ARGUMENTS

    name = str: name of volume object
    '''
    from pymol import gui

    qt_window = not int(_noqt) and gui.get_qtwindow()
    app = gui.get_pmgapp()

    if qt_window:
        from pmg_qt import volume as volume_qt

        @app.execute
        def _():
            try:
                panel = _volume_windows_qt[name]
            except LookupError:
                panel = volume_qt.VolumePanel(qt_window, name, _self=_self)
                _volume_windows_qt[name] = panel
            panel.show()
            panel.raise_()

        return

    # Tk fallback
    from pmg_tk import volume
    import tkinter as Tkinter

    def func():
        try:
            window = _volume_windows[name]
            window.lift()
        except (LookupError, Tkinter.TclError):
            window = Tkinter.Toplevel(app.root)
            window.title('Volume Panel for "%s"' % name)
            window.panel = volume.VolumePanel(window, name, _self=_self)
            window.panel.pack()
            _volume_windows[name] = window
    app.execute(func)

### utility functions

def ramp_to_colors(ramp, vmin=None, vmax=None, ncolors=360):
    '''
    Get the interpolated color array for the given ramp and data range
    '''
    ramp = ramp_expand(ramp)

    if vmin is None: vmin = ramp[0]
    if vmax is None: vmax = ramp[-5]

    colors = [(0.0, 0.0, 0.0, 0.0)] * ncolors
    upper = None

    for i in range(len(ramp) / 5):
        v = ramp[i * 5]
        lower = upper
        upper = int(ncolors * (v - vmin) / (vmax - vmin))

        if i == 0:
            continue

        mixcincr = 1.0 / (upper - lower)
        mixc = 1.0 + mixcincr

        for j in range(lower, upper):
            mixc -= mixcincr
            if j < 0 or j >= ncolors:
                continue
            colors[j] = [
                    ramp[i * 5 - 1 + k + 1] * mixc + \
                    ramp[i * 5     + k + 1] * (1.0 - mixc)
                    for k in range(4)]

    return colors

def ramp_expand(ramp, _self=cmd):
    '''
    Takes a list or space separated string of (v,r,g,b,a) or (v,colorname,a)
    specs and flattens it to a list of floats.

    >>> print ramp_expand('1.2 blue 0.3 4.5 red .7')
    [1.2, 0.0, 0.0, 1.0, 0.3, 4.5, 1.0, 0.0, 0.0, 0.7]

    @rtype: list
    '''
    if isinstance(ramp, str):
        if ramp[:1] in '([':
            ramp = _self.safe_eval(ramp)
        else:
            ramp = ramp.split()
    it = flatiter(ramp)
    ramp = []
    for s in it:
        try:
            c = next(it)
            try:
                r = float(c)
                if r > 1.0:
                    # support for numeric color index except 0+1 (white+black)
                    raise ValueError
            except ValueError:
                r, g, b = _self.get_color_tuple(c)
            else:
                g, b = float(next(it)), float(next(it))
            a = float(next(it))
        except StopIteration:
            raise ValueError("malformed color ramp")
        ramp.extend([float(s), r, g, b, a])
    return ramp

def flatiter(x):
    '''
    Flat iterator over nested list-like structures (does not flatten strings).

    >>> list(flatiter([1,2,(3,4),5,("foo", (7,8), "bar"),10]))
    [1, 2, 3, 4, 5, 'foo', 7, 8, 'bar', 10]

    Details: Flattens all elements with an __iter__ method. Sequences with
    __len__ and __getitem__ are also iterable, but are not flattened by this
    function.

    @type x: iterable
    @rtype: generator
    '''
    x = [iter(x)]
    while x:
        try:
            e = next(x[-1])
        except:
            x.pop()
            continue
        if hasattr(e, '__iter__') and not isinstance(e, str):
            x.append(iter(e))
            continue
        yield e

def flatlist(x):
    '''
    Same as list(flatiter(x)) but faster.
    '''
    r = []
    def loop(x):
        for e in x:
            if hasattr(e, '__iter__') and not isinstance(e, str):
                loop(e)
            else:
                r.append(e)
    loop(x)
    return r



import math

import re
import pymol
cmd = __import__("sys").modules["pymol.cmd"]
from . import setting
from . import parsing
import threading

from .cmd import DEFAULT_ERROR, DEFAULT_SUCCESS, _raising, \
     is_list,  is_string, is_error

QuietException = parsing.QuietException

_prefix = "_tmp_editor"
tmp_wild = _prefix + "*"
tmp_editor = _prefix + "0"
tmp_connect = _prefix + "_con"
tmp_domain = _prefix + "_dom"
tmp1 = _prefix + "1"
tmp2 = _prefix + "2"
tmp3 = _prefix + "3"
tmp4 = _prefix + "4"

# routines to assist in molecular editing

class undocontext:
    def __init__(self, cmd, sele):
        # not implemented in open-source
        pass

    def __enter__(self):
        # not implemented in open-source
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        # not implemented in open-source
        pass

def attach_fragment(selection,fragment,hydrogen,anchor,*,_self=cmd):
    '''
ARGUMENTS

    selection = str: Name of a single-atom selection. If no such named
    selection exists, then create a new object with name `fragment`.

    fragment = str: fragment name to load from fragment library

    hydrogen = int: atom ID in fragment to fuse

    anchor = int: (unused)
    '''
    remove_hydrogens = _self.get_setting_boolean("auto_remove_hydrogens")

    if selection not in _self.get_names("selections"):
        if fragment in _self.get_names("objects"):
            raise pymol.CmdException("an object with that name already exists")

        _self.fragment(fragment)

        if remove_hydrogens:
            _self.remove(f"(hydro and {fragment})")
    else:
        fragment_label = _self.get_unused_name(_prefix + "_attach_fragment")
        _self.fragment(fragment, fragment_label, origin=0)

        try:
            _self.fuse(f"{fragment_label} and id {hydrogen}", f"({selection})", 1)

            if remove_hydrogens:
                _self.remove("(hydro and pkmol)")
            elif _self.count_atoms('hydro and (neighbor pk2)'):
                _self.h_fill()
        finally:
            _self.delete(fragment_label)

def combine_fragment(selection,fragment,hydrogen,anchor,_self=cmd):
    with undocontext(_self, selection):
        _self.fragment(fragment,tmp_editor)
        try:
            if _self.get_setting_boolean("auto_remove_hydrogens"):
                _self.remove("(hydro and ?%s)" % tmp_editor)
            _self.fuse("?%s" % tmp_editor, "(%s)" % selection, 3)
        finally:
            _self.delete(tmp_editor)

def attach_amino_acid(selection,amino_acid,center=0,animate=-1,object="",hydro=-1,ss=-1,_self=cmd):
    '''
ARGUMENTS

    selection = str: named selection of single N or C atom

    amino_acid = str: fragment name to load from fragment library

    center = bool: center on new terminus (pk1)

    animate = int: animate centering

    object = str: name of new object (if selection is none)

    hydro = int (-1/0/1): keep hydrogens

    ss = int: Secondary structure 1=alpha helix, 2=antiparallel beta, 3=parallel beta, 4=flat
    '''
    r = DEFAULT_SUCCESS
    ss = int(ss)
    center = int(center)

    if hydro<0:
        hydro = not int(_self.get_setting_boolean("auto_remove_hydrogens"))
    if (selection not in _self.get_names('all')
            if selection == 'pk1' # legacy, calling functions should pass '?pk1'
            else _self.count_atoms(selection) == 0):
        if object == "":
            object = amino_acid
        # create new object
        if amino_acid in _self.get_names("objects"):
            print("Error: an object with than name already exists")
            raise QuietException
        r = _self.fragment(amino_acid,object)
        if not hydro:
            _self.remove("(hydro and %s)"%object)
        if _self.count_atoms("((%s) and name C)"%object):
            _self.edit("((%s) and name C)"%object)
        elif _self.count_atoms("((%s) and name N)"%object):
            _self.edit("((%s) and name N)"%object)
    elif _self.select(tmp_connect,"(%s) & elem N,C"%selection) != 1:
        print("Error: invalid connection point: must be one atom, name N or C.")
        _self.delete(tmp_wild)
        raise QuietException
    elif amino_acid in ["nhh","nme"] and _self.select(tmp_connect,"(%s) & elem C"%selection) != 1:
        print("Error: invalid connection point: must be C for residue '%s'"%(amino_acid))
        _self.delete(tmp_wild)
        raise QuietException
    elif amino_acid in ["ace"] and _self.select(tmp_connect,"(%s) & elem N"%selection) != 1:
        print("Error: invalid connection point: must be N for residue '%s'"%(amino_acid))
        _self.delete(tmp_wild)
        raise QuietException
    else:
        if ss<0:
            ss = _self.get_setting_int("secondary_structure")
        if ss:
            if ss==1: # helix
                phi=-57.0
                psi=-47.0
            elif ss==2: # antipara-beta
                phi=-139.0
                psi=135.0
            elif ss==3: # para-beta
                phi=-119.0
                psi=113.0
            else:
                phi=180.0
                psi=180.0
        _self.fragment(amino_acid,tmp_editor, origin=0)
        if _self.count_atoms("elem N",domain=tmp_connect):
            tmp = [ None ]
            _self.iterate(tmp_connect,"tmp[0]=resv", space={ 'tmp' : tmp })
            tmp[0] = str(tmp[0]-1) # counting down
            _self.alter(tmp_editor,"resi=tmp[0]",space={ 'tmp' : tmp})
            _self.set_geometry(tmp_connect, 3, 3) # make nitrogen planar
            _self.fuse("(%s and name C)"%(tmp_editor),tmp_connect,2)
            _self.select(tmp_domain, "byresi (pk1 | pk2)")

            if not hydro:
                _self.remove("(pkmol and hydro)")

            if ((_self.select(tmp1,"?pk1",domain=tmp_domain)==1) and
                (_self.select(tmp2,"?pk2",domain=tmp_domain)==1)):

                if ((_self.select(tmp3,"(name CA,CH3 & nbr. ?pk1)",domain=tmp_domain)==1) and
                    (_self.select(tmp4,"(name CA,CH3 & nbr. ?pk2)",domain=tmp_domain)==1)):
                    _self.set_dihedral(tmp4,tmp2,tmp1,tmp3,180.0)

                if hydro:
                    _self.h_fix(tmp2) # fix hydrogen position

                if ss:
                    if amino_acid[0:3]!='pro':
                        if ((_self.select(tmp4,
                                          "(!(resn PRO) & name C  & nbr. (name CA & nbr. "+tmp2+"))",
                                          domain=tmp_domain)==1) and
                            (_self.select(tmp3,
                                          "(!(resn PRO) & name CA & nbr. "+tmp2+")",
                                          domain=tmp_domain)==1)):
                            _self.set_dihedral( # PHI
                                tmp4, # C
                                tmp3, # CA
                                tmp2, # N
                                tmp1, # C
                                phi)

                    if ((_self.select(tmp4,"(name N & nbr. (name CA & nbr. "+tmp1+"))",
                                      domain=tmp_domain)==1) and
                        (_self.select(tmp3,"(name CA & nbr. "+tmp1+")",domain=tmp_domain)==1)):
                        _self.set_dihedral( # PSI (n-1)
                            tmp2, # N
                            tmp1, # C
                            tmp3, # CA
                            tmp4, # N
                            psi)

            sele = ("(name N & (byres nbr. %s) &! (byres %s))"% (tmp_connect,tmp_connect))
            if _self.select(tmp1,sele,domain=tmp_domain):
                _self.edit(tmp1)
                if center:
                    _self.center(tmp1,animate=animate)
        elif _self.count_atoms("elem C",domain=tmp_connect): # forward
            tmp = [ None ]
            _self.iterate(tmp_connect,"tmp[0]=resv", space={ 'tmp' : tmp })
            tmp[0] = str(tmp[0]+1) # counting up
            _self.alter(tmp_editor,"resi=tmp[0]",space={ 'tmp' : tmp})
            _self.set_geometry(tmp_editor + " & name N", 3, 3) # make nitrogen planar
            _self.fuse("(%s and name N)"%tmp_editor,tmp_connect,2)
            _self.select(tmp_domain, "byresi (pk1 | pk2)")

            if not hydro:
                _self.remove("(pkmol and hydro)")

            if (( _self.select(tmp1,"?pk1",domain=tmp_domain)==1) and
                ( _self.select(tmp2,"?pk2",domain=tmp_domain)==1)):

                if ((_self.select(tmp3,"(name CA,CH3 & nbr. ?pk1)",domain=tmp_domain)==1) and
                    (_self.select(tmp4,"(name CA,CH3 & nbr. ?pk2)",domain=tmp_domain)==1)):
                    _self.set_dihedral(tmp4,tmp2,tmp1,tmp3,180.0)
                if hydro:
                    _self.h_fix("pk1") # fix hydrogen position
                if ss:
                    if hydro and amino_acid[0:3]=='nhh': # fix amide hydrogens
                        if ((_self.select(tmp3,"(name H1 & nbr. "+tmp1+")",domain=tmp_domain)==1) and
                            (_self.select(tmp4,"(name O & nbr. "+tmp2+")",domain=tmp_domain)==1)):
                            _self.set_dihedral(
                                tmp4, # O
                                tmp2, # C
                                tmp1, # N
                                tmp3, # H1
                                180)
                    if amino_acid[0:3]!='pro':
                        if ((_self.select(tmp3,"(name CA & nbr. "+tmp1+")",domain=tmp_domain)==1) and
                            (_self.select(tmp4,"(name C & nbr. (name CA & nbr. "+tmp1+"))",domain=tmp_domain)==1)):
                            _self.set_dihedral( # PHI
                                tmp2, # C
                                tmp1, # N
                                tmp3, # CA
                                tmp4, # C
                                phi)
                    if ((_self.select(tmp3,"(name CA & nbr. "+tmp2+")",domain=tmp_domain)==1) and
                        (_self.select(tmp4,"(name N & nbr. (name CA & nbr. "+tmp2+"))",domain=tmp_domain)==1)):
                        _self.set_dihedral( # PSI (n-1)
                            tmp4, # N
                            tmp3, # CA
                            tmp2, # C
                            tmp1, # N
                            psi)
            sele = ("(name C & (byres nbr. %s) & !(byres %s))"% (tmp_connect,tmp_connect))
            if _self.select(tmp1,sele,domain=tmp_domain):
                _self.edit(tmp1)
                if center:
                    _self.center(tmp1,animate=animate)
            else:
                _self.unpick()
        elif _self.count_atoms("((%s) and elem H)"%selection):
            print("Error: please pick a nitrogen or carbonyl carbon to grow from.")
            _self.delete(tmp_wild)
            raise QuietException
        else:
            print("Error: unable to attach fragment.")
            _self.delete(tmp_wild)
            raise QuietException
    _self.delete(tmp_wild)

    return r

_aa_codes =  {
    'A' : 'ala',
    'B' : 'ace',
    'C' : 'cys',
    'D' : 'asp',
    'E' : 'glu',
    'F' : 'phe',
    'G' : 'gly',
    'H' : 'his',
    'I' : 'ile',
    'K' : 'lys',
    'L' : 'leu',
    'M' : 'met',
    'N' : 'asn',
    'P' : 'pro',
    'Q' : 'gln',
    'R' : 'arg',
    'S' : 'ser',
    'T' : 'thr',
    'V' : 'val',
    'W' : 'trp',
    'Y' : 'tyr',
    'Z' : 'nme',
    }

_fab_codes = {
    'peptide' : _aa_codes,
    }

_pure_number = re.compile("[0-9]+")

def _fab(input,name,mode,resi,chain,segi,state,dir,hydro,ss,quiet,_self=cmd):
    r = DEFAULT_ERROR
    code = _fab_codes.get(mode,None)
    quiet = int(quiet)
    resi = int(resi)
    state = int(state)
    dir = int(dir)
    hydro = int(hydro)

    if hydro < 0:
        hydro = not _self.get_setting_boolean("auto_remove_hydrogens")

    seq_len = 0
    if (mode == 'peptide') and is_string(input):
        # '123/ ADC B/234/ AFCD' to [ '123/','A','D','C','B/234/','F','C','D' ]
        frags = input.split()
        input = []
        for frag in frags:
            if '/' in frag:
                input.append(frag)
            else:
                seq_len = seq_len + len(frag)
                input.extend(list(frag))
                input.append("/") # breaks chain
    if name is None:
        name = _self.get_unused_name("obj")
    elif name in _self.get_names():
        _self.delete(name)

    if mode in [ 'peptide' ]:  # polymers
        if (seq_len>99) and not quiet:
            print(" Generating a %d residue peptide from sequence..."%seq_len)
        input.reverse()
        sequence = input
        if code is not None:
            while len(sequence):
                while len(sequence) and '/' in sequence[-1]:
                    part = sequence.pop().split('/')
                    if len(part)>1:
                        if len(part[-2]):
                            resi = int(part[-2])
                    if len(part)>2:
                        chain = part[-3]
                    if len(part)>3:
                        segi = part[-4]
                if len(sequence) and not _self.count_atoms("?pk1"): # new polymer segment
                    tmp_obj = _self.get_unused_name()
                    first = sequence.pop()
                    _self.fragment(code[first], tmp_obj)
                    if not hydro:
                        _self.remove(tmp_obj + ' and hydro')
                    _self.alter(tmp_obj,'resi="""%s""";chain="""%s""";segi="""%s"""'%(resi,chain,segi))
                    _self.create(name,tmp_obj+" or ?"+name,1,1,zoom=0)
                    tmp_sel = _self.get_unused_name()
                    if mode == 'peptide':
                        if dir>0:
                            _self.select(tmp_sel,"name C and "+tmp_obj)
                            resi = resi + 1
                        else:
                            _self.select(tmp_sel,"name N and "+tmp_obj)
                            resi = resi - 1
                    _self.edit(name+" in "+tmp_sel) # set the editor's pk1 selection
                    _self.delete(tmp_sel+" "+tmp_obj)
                if mode == 'peptide':
                    while len(sequence):
                        if '/' in sequence[-1]:
                            _self.unpick() # break chain at this point
                            break
                        if not _self.count_atoms("?pk1"):
                            break
                        else:
                            attach_amino_acid("pk1",code[sequence.pop()],animate=0,ss=ss,hydro=hydro,_self=_self)
                            if dir>0:
                                resi = resi + 1
                            else:
                                resi = resi - 1
    if not len(sequence):
        r = DEFAULT_SUCCESS

    if _self.get_setting_int('auto_zoom'):
        _self.zoom(name)

    return r

_threeNA_to_OneNA = { "atp" : "A", "ctp" : "C", "gtp" : "G", "ttp" : "T", "utp" : "U"}
_oneNA_to_threeNA = { "A" : "atp", "C" : "ctp", "G" : "gtp", "T" : "ttp", "U" : "utp"}

_base_pair = { "DNA" : {"atp" : "ttp", "ctp" : "gtp", "gtp" : "ctp", "ttp" : "atp" },
               "RNA" : {"atp" : "utp", "ctp" : "gtp", "gtp" : "ctp", "utp" : "atp" }}

_oneNA_base_pair = {"DA" : "DT", "DC" : "DG", "DG" : "DC", "DT" : "DA", "A" : "U",
                    "C" : "G", "G" : "C", "U" : "A"}

def iterate_to_list(selection: str, expression: str, *, _self=cmd):
    outlist = []
    _self.iterate(selection,f"outlist.append(({expression}))", space={"outlist":outlist})
    return outlist

def rename_three_to_one(nuc_acid, sele, nuc_type, *, _self=cmd):
    """
    Renames nucleobase from 3-letter to 1-letter representation

    :param nuc_acid: (str) 3-letter nucleic acid representation
    :param sele: (str) selection of nucleic acid to rename
    :param nuc_type: (str) "DNA" or "RNA"
    """
    new_name = _threeNA_to_OneNA[nuc_acid]
    if nuc_type == "DNA":
        new_name = "D" + new_name
    _self.alter(sele, f"resn='{new_name}'")

def fit_sugars(mobile, target, *, _self=cmd):
    """
    Fits appending base pairs to form appropriate hydrogen bond

    :param mobile: (str) selection for the sense (main) strand
    :param target: (str) selection for the antisense (opposing) strand
    """
    try:
        _self.pair_fit(f"{mobile} & name C1'",
                       f"{target} & name C1'",
                       f"{mobile} & name C2'",
                       f"{target} & name C2'",
                       f"{mobile} & name C3'",
                       f"{target} & name C3'",
                       f"{mobile} & name C4'",
                       f"{target} & name C4'",
                       f"{mobile} & name O4'",
                       f"{target} & name O4'", quiet=1)
    except:
        _self.delete(tmp_wild)
        raise pymol.CmdException("Something went wrong when fitting the new residue.")

def fit_DS_fragment(mobile_A, target_A, mobile_B, target_B, *, _self=cmd):
    """
    Fits dummy fragment to the detected structure using atoms
    on both stands for a more accurate alignment.

    :param mobile_A: (str) selection for the base being created and attached
    :param target_A: (str) selection for the base selected to build on
    :param mobile_B: (str) selection for the opposing base being created
    :param target_B: (str) selection for the detected opposing base
     """
    try:
        _self.pair_fit(f"{mobile_A} & name C1'",
                       f"{target_A} & name C1'",
                       f"{mobile_A} & name C2'",
                       f"{target_A} & name C2'",
                       f"{mobile_A} & name C5'",
                       f"{target_A} & name C5'",
                       f"{mobile_A} & name O4'",
                       f"{target_A} & name O4'",
                       f"{mobile_A} & name O3'",
                       f"{target_A} & name O3'",
                       f"{mobile_A} & name P",
                       f"{target_A} & name P",
                       f"{mobile_B} & name O3'",
                       f"{target_B} & name O3'",
                       f"{mobile_B} & name C1'",
                       f"{target_B} & name C1'",
                       f"{mobile_B} & name C2'",
                       f"{target_B} & name C2'",
                       f"{mobile_B} & name C5'",
                       f"{target_B} & name C5'",
                       f"{mobile_B} & name P",
                       f"{target_B} & name P",
                       f"{mobile_B} & name O4'",
                       f"{target_B} & name O4'", quiet=1)
    except:
        _self.delete(tmp_wild)
        raise pymol.CmdException("Something went wrong when fitting the new residue.")

def add2pO(domain, nuc_acid, resv, *, _self=cmd):
    if nuc_acid == "utp": #utp comes with O2'
        return
    c_2p = _prefix + "_c2p"
    _self.select(c_2p, "%s & resi \\%i & name %s" % (domain, resv, "C2'"))
    _self.unpick()
    _self.edit(c_2p)
    _self.attach("O", 4, 4)
    _self.unpick()
    _self.alter("(byres %s) & resi \\%i & name O01" % (domain, resv), "name=\"O2'\"")

def move_atom_in_res(atom, dummy_res, new_res, twist, rise, *, _self=cmd):
    prev_coords = _self.get_coords(f"{dummy_res} & name {atom}", state=1)
    curr_coords = _self.get_coords(f"{new_res} & name {atom}", state=1)

    if curr_coords is None or prev_coords is None:
        return
    curr_coord = curr_coords[0]
    prev_coord = prev_coords[0]

    r = math.sqrt(prev_coord[0] ** 2 +
                  prev_coord[1] ** 2)
    old_phi = math.degrees(math.atan2(prev_coord[1],prev_coord[0]))
    phi = old_phi - twist
    phi = math.radians(phi)
    new_pos = [r * math.cos(phi),
               r * math.sin(phi),
               prev_coord[2] - rise]

    trans = list(new_pos - curr_coord)
    _self.translate(trans, f"{new_res} & name {atom}", camera=0)

def move_new_res(frag_string, full_frag, old, old_oppo, double_stranded_bool=False, form="B", chain="A", antisense=False, *, _self=cmd):
    """
    Attaches new residue (or pair) onto current nucleotide chain

    :param frag_string: (str) Name of appending nucleic acid or pair 
    :param full_frag: (str) Selection of the created fragment 
    :param old: (str) Selection of previous residue
    :param old_oppo: (str) Selection of previous opposing residue
    :param double_stranded_bool: (bool) Flag represing if double helix was detected
    :param antisense (bool) Flag for antisense
    :param form: (str) DNA form ('A'/'B')
    """
    if form == 'B':
        twist = -36.0
        rise = -3.375
    elif form == 'A':
        twist = -32.7
        rise = -2.548
    else:
        raise ValueError("Form not recognized")

    rise = rise if antisense else -rise
    twist = twist if antisense else -twist

    dummy_fragment = _prefix + "_dummyfrag"
    dummy_res_A = _prefix + "_dummyresA"
    dummy_res_B = _prefix + "_dummyresB"
    new_fragment = _prefix + "_newfrag"
    new_res_A = _prefix + "_newresA"
    new_res_B = _prefix + "_newresB"

    _self.select(new_fragment, f"{full_frag}")
    _self.select(new_res_A,f"{full_frag} and chain A")
    _self.select(new_res_B,f"{full_frag} and chain B")

    _self.fragment(frag_string, dummy_fragment, origin=0)
    _self.select(dummy_res_A, f"{dummy_fragment} and chain A")
    _self.select(dummy_res_B, f"{dummy_fragment} and chain B")

    if old_oppo == "none":
        # This is the case where a single residue is being added
        _self.select(dummy_res_A, f"{dummy_fragment}")
        _self.select(new_res_A, f"{full_frag}")

    atoms_A = iterate_to_list(dummy_res_A, "name")
    atoms_B = iterate_to_list(dummy_res_B, "name")

    #A new base is created by copying the coordinates of the previous
    #base and doing a cylindrical rotation (phi degrees) and a translation
    #down the z-axis by the rise amount
    for atom in atoms_A:
        move_atom_in_res(atom, dummy_res_A, new_res_A, twist, rise)
    for atom in atoms_B:
        move_atom_in_res(atom, dummy_res_B, new_res_B, twist, rise)
    if double_stranded_bool == True:
        fit_DS_fragment(dummy_res_A, old, dummy_res_B, old_oppo)
        orient_flag = check_dummy_oriention(old,dummy_res_A)
        if orient_flag == 0:
            fit_sugars(dummy_res_A,old)
    elif double_stranded_bool == False:
        fit_sugars(dummy_res_A, old)
    else:
        raise pymol.CmdException("Double stranded bool was not provided to move_new_res")

    dummy_fragment_transform = _self.get_object_matrix(dummy_fragment)
    _self.transform_object(full_frag, dummy_fragment_transform)
    _self.delete(dummy_fragment)

def check_dummy_oriention(old, dummy_res_A, *, _self=cmd):
    dummy_orient = _prefix + "_dummy_orient"
    orient_flag = 0
    orient_flag = _self.select(dummy_orient, f"({old} & name O4') within 1.0 of ({dummy_res_A} & name O4')")

    return orient_flag

class NascentNucAcidInfo:
    def __init__(self, fragment_name, nuc_acid, nuc_type, form, dbl_helix):
        self.fragment_name = fragment_name
        self.nuc_acid = nuc_acid
        self.nuc_type = nuc_type
        self.form = form
        self.dbl_helix = dbl_helix

def attach_O5_phosphate(_self=cmd):
    if "pk1" not in _self.get_names("selections"):
        raise pymol.CmdException("Selection must be pk1 to attach O5' phosphate")

    print("This building selection has an unphosphorylated O5' end.")
    attach_fragment("pk1","phosphite",4,0)
    # Initailize selection strings
    P_center = _prefix + "_P_center"
    H_extra = _prefix + "_H_extra"
    O_one = _prefix + "_O_one"
    O_two = _prefix + "_O_two"
    O_three = _prefix + "_O_three"

    # Selection
    _self.select(P_center,"n. P01")
    _self.select(H_extra, f"h. and bound_to {P_center} or n. H02")
    _self.select(O_one, f"n. O01 and bound_to {P_center}")
    _self.select(O_two, f"n. O02 and bound_to {P_center}")
    _self.select(O_three, f"n. O03 and bound_to {P_center}")

    # Removing unnecessary atoms
    _self.remove(H_extra)
    _self.remove(O_one)

    # Fix bonding
    _self.unbond(P_center,O_three)
    _self.bond(P_center,O_three,1)
    _self.unbond(P_center,O_two)
    _self.bond(P_center,O_two,2)

    # Rename P correctly
    _self.alter(P_center,"name = 'P'")

    # Set Pk1 correctly
    _self.select("pk1",P_center)

def check_DNA_base_pair(sele_oppo_atom, selection, *, _self=cmd):
    base_pair_dist = _prefix + "_base_pair_dist"
    base_pair_bool = 0
    tmp_last_resn = iterate_to_list(selection,"resn")
    tmp_last_resn_oppo = iterate_to_list(sele_oppo_atom,"resn")

    if len(tmp_last_resn_oppo) != 0:
        last_resn = str(tmp_last_resn[0])
        last_resn_oppo = str(tmp_last_resn_oppo[0])

        if (_oneNA_base_pair[last_resn] == last_resn_oppo and
            _self.select(base_pair_dist, f"(byres {sele_oppo_atom}) within 3.5 of (byres {selection})") != 0):
                base_pair_bool = 1
        else:
            base_pair_bool = 0
    else:
        print("check_DNA_base_pair has no opposing residue to check")
    return base_pair_bool

def get_chains_oppo (chain, tmp_connect, *, _self=cmd):
    models = iterate_to_list(tmp_connect,"model")

    close_chains = []
    close_chains = _self.get_chains(f"({models[0]}) within 15.0 of {tmp_connect}")
    close_chains = [c for c in close_chains if c != chain]

    return close_chains

def get_new_chain (chain, tmp_connect, *, _self=cmd):
    models = iterate_to_list(tmp_connect,"model")
    model_chains = _self.get_chains(models[0])
    search_chain_flag = 0

    if len(model_chains) != 0:
        last_chain = f"{model_chains[-1]}"
        last_chain_front = last_chain[:-1]
        last_chain_back = last_chain[-1]
        if last_chain_back != 'Z' and last_chain_back != 'z':
            new_chain_back = chr(ord(last_chain_back)+1)
        elif last_chain_back == 'Z':
            new_chain_back = "ZA"
            print("Z chain was detected. New chain will append A")
        else:
            new_chain_back = "za"
            print("z chain was detected. New chain will append a")
    else:
        new_chain_back = "A"

    chain_oppo = last_chain_front + new_chain_back
    return chain_oppo

def check_valid_attachment(nascent, atom_selection_name, selection, resv, *, _self=cmd):
    atom_selection_name_partner = "O3'" if atom_selection_name == "P" else "P"
    atom_sele = _prefix + "atom_sele"
    _self.select(atom_sele, f"{selection}")
    bound = _prefix + "atom_sele_bound"
    if _self.count_atoms(f"(bound_to {atom_sele}) & name {atom_selection_name_partner}") != 0:
        _self.delete(tmp_wild)
        raise pymol.CmdException(f"{atom_selection_name} already bonded!")

def bond_single_stranded(tmp_editor, object, chain, resv, last_resi_sele, atom_selection_name, atom_name_oppo, *, _self=cmd):
    """
    Forms a bond between the last atoms on selected structure and the newly created fragment
    :param tmp_editor: (str) Object representing the newly created fragment
    :param object: (str) object/model name of the selected structure
    :param chain: (str) Chain ID
    :param resv: (int)
    :param last_resi_sele: (str) The selection string of the selected residue
    :param atom_selection_name: (str) Name of the atom selected
    :param atom_name_oppo: (str) Name of the corresponding opposing atom
    """
    object_fuse = _prefix + f"_{chain}_fuse"
    object_connect = _prefix + f"_{chain}_con"

    print("The program did not detect a double stranded structure, so the opposing residue will not be attached.")

    # Select and fuse
    _self.select(object_fuse, f"{last_resi_sele} & name {atom_selection_name}")
    _self.fuse(f"{tmp_editor} & chain {chain} & name {atom_name_oppo}", object_fuse, mode=3)
    _self.select(object_connect, f"{last_resi_sele} & name {atom_selection_name} & chain {chain}")

    # Target is on the new fragment
    object_bond_target = _prefix + f"_{chain}_con_target"
    if (_self.select(object_bond_target, f"{object} & resi \\{resv} & name {atom_name_oppo} & chain {chain}") == 1):
        bond_dist = _prefix + "_bond_dist"
        if (_self.select(bond_dist, f"{object_connect} within 3.0 of {object_bond_target}") != 0):
            _self.bond(object_connect, object_bond_target)
        else:
            print("Identified bond targets were too far apart, so this will not be bound")
    else:
        print("More than one bond target was identified, so this will not be bound")

def bond_double_stranded(tmp_editor, object, chain, chain_oppo, resv, resv_oppo, last_resi_sele, prev_oppo_res, atom_selection_name,
                         atom_name_oppo, *, _self=cmd):
    """
    Forms a bond between the last atoms on selected structure and the newly created fragment
    :param tmp_editor: (str) Object representing the newly created fragment
    :param object: (str) object/model name of the selected structure
    :param chain: (str) Chain ID
    :param chain_oppo: (str) Opposing chain ID
    :param resv: (int)
    :param resv_oppo: (int)
    :param last_resi_sele: (str) The selection string of the selected residue
    :param prev_oppo_res: (str) The selection string of the previous opposing residue
    :param atom_selection_name: (str) Name of the atom selected
    :param atom_name_oppo: (str) Name of the corresponding opposing atom
    """
    object_fuse = _prefix + f"_{chain}_fuse"
    object_oppo_fuse = _prefix + f"_oppo_{chain_oppo}_fuse"
    object_connect = _prefix + f"_{chain}_con"
    object_oppo_connect = _prefix + f"_oppo_{chain_oppo}_con"

    # Target is on the new fragment
    object_bond_target = _prefix + f"_{chain}_con_target"
    object_oppo_bond_target = _prefix + f"_oppo_{chain_oppo}_con_target"

    # Select and fuse
    _self.select(object_fuse, f"{last_resi_sele} & name {atom_selection_name}")
    _self.select(object_oppo_fuse, f"{prev_oppo_res} & name {atom_name_oppo}")
    _self.fuse(f"{tmp_editor} & chain {chain} & name {atom_name_oppo}", object_fuse, mode=3)

    if ((_self.select(object_bond_target, f"{object} & resi \\{resv} & name {atom_name_oppo} & chain {chain}") == 1) and
        (_self.select(object_connect, f"{last_resi_sele} & name {atom_selection_name} & chain {chain}") == 1)):
        bond_dist = _prefix + "_bond_dist"
        if (_self.select(bond_dist, f"{object_connect} within 3.0 of {object_bond_target}") != 0):
            _self.bond(object_connect, object_bond_target)
        else:
            print("Identified bond targets were too far apart, so this will not be bound")
    else:
        print("More than one bond target was found on selected chain, so this will not be bound.")

    if ((_self.select(object_oppo_bond_target, f"{object} & resi \\{resv_oppo} & name {atom_selection_name} & chain {chain_oppo}") == 1) and
        (_self.select(object_oppo_connect, f"{prev_oppo_res} & name {atom_name_oppo} & chain {chain_oppo}") == 1)):
        bond_dist = _prefix + "_bond_dist"
        if (_self.select(bond_dist, f"{object_oppo_connect} within 3.0 of {object_oppo_bond_target}") != 0):
            _self.bond(object_oppo_connect, object_oppo_bond_target)
        else:
            print("Identified bond targets were too far apart, so this will not be bound")
    else:
        print("More than one bond target was found on opposing chain, so this will not be bound.")

def attach_nuc_acid(selection, nuc_acid, nuc_type, object= "", form ="B",
                        dbl_helix=True, *, _self=cmd):
    """
    Creates new nuc acid attached to existing PDB structure 
    :param selection: (str) selection of picked nascent chain (or nothing)
    :param nuc_acid: (str) appending nucleic acid
    :param nuc_type: (str) sugar type of nucleic acid
    :param object: (str) name of appending nucleobase
    :param form: (str) DNA structure form: A, B, or Z
    :param dbl_helix: (bool) flag for double-strandedness
    """
    original_sele = _prefix + "_original_sele"
    _self.select(original_sele,selection)

    if nuc_type == "RNA" and form != 'A':
        form = 'A'
        dbl_helix = False

    nascent = NascentNucAcidInfo(nuc_acid + form, nuc_acid, nuc_type, form, dbl_helix)
    nuc_acid_partner_temp = _base_pair[nuc_type][nuc_acid].lower()
    nascent_partner =  NascentNucAcidInfo(nuc_acid_partner_temp + form, nuc_acid_partner_temp,
                                          nuc_type, form, dbl_helix)

    if _self.cmd.count_atoms(selection) == 0:
        if object == "":
            object = nuc_acid

        if dbl_helix:
            frag_string = nascent.nuc_acid + "_"+ _base_pair["DNA"][nascent.nuc_acid] + nascent.form
            _self.fragment(frag_string,object)
        elif not dbl_helix:
            _self.fragment(nascent.fragment_name, object, origin=0)
            _self.alter(object, f"segi='A';chain='A';resv=1")
            rename_three_to_one(nascent.nuc_acid, object, nascent.nuc_type)

        if nascent.nuc_type == "RNA":
            add2pO(object, nascent.nuc_acid, 1)

        if _self.count_atoms(f"{object} & segi A & name P"):
            _self.edit(f"{object} & segi A & name P")
        elif _self.count_atoms(f"{object} & segi A & name O3'"):
            _self.edit(f"{object} & segi A & name O3'")
        _self.edit(f"{object} & segi A & name O3'")
        _self.select("pk1", f"{object} & name O3' & chain A")
    elif _self.select(tmp_connect, f'{selection}') == 1:
        chain, name, object = iterate_to_list(selection,"chain,name,model")[0]

        if name == "O5'":
            attach_O5_phosphate()
            name = "P"
            # FIXME Don't use `selection` as the name, or ensure that it's
            # indeed just a name and not a selection expression.
            _self.select(selection,f"name P & bound_to {original_sele}")
            _self.delete("_pkdihe")
            _self.select(tmp_connect,f"{selection}")

        if name == "P" or name == "O3'":
            extend_nuc_acid(nascent, nascent_partner, selection, object, name, chain, _self=_self)

        _self.select("pk1","_tmp_editor_new_selection")
    else:
        _self.delete(tmp_wild)
        raise pymol.CmdException("invalid connection point: must be one atom, name O3' or P")

    _self.show("cartoon", f"byobject {selection}")
    _self.delete(tmp_wild)

def extend_nuc_acid(nascent, nascent_partner, selection,
              object, atom_selection_name, chain = 'A', *, _self=cmd):
    """
    Creates new nuc acid (or pair) or attaches to PDB chain
    :param nascent: (NascentNucAcidInfo) appending nucleic acid
    :param nascent_partner: (NascentNucAcidInfo) partner of appending nucleic acid
    :param selection: (str) selection string of selected residue
    :param object: (str) object of selected residue
    :param atom_selection_name: (str) O3' or P
    :param chain: (str) chain ID

    FIXME Eliminate `tmp_connect` pre-condition (or at least document it
    properly). Looks like `tmp_connect` must be the same as `selection`?
    """
    # Making a temporary selection
    original_sele = _prefix + "_original_sele"
    _self.select(original_sele,selection)

    # Alter the segi to match the chain selection
    chain_sele = "_chain_sele"
    _self.select(chain_sele, f"chain {chain}")
    _self.alter(chain_sele,f"segi = '{chain}'")
    _self.delete(chain_sele)

    if not nascent.dbl_helix:
        frag_string = nascent.fragment_name
        _self.fragment(frag_string, tmp_editor, origin=0)
        rename_three_to_one(nascent.nuc_acid, tmp_editor, nascent.nuc_type)
    elif nascent.dbl_helix:
        frag_string = nascent.nuc_acid + "_"+ _base_pair["DNA"][nascent.nuc_acid] + nascent.form
        _self.fragment(frag_string, tmp_editor, origin=0)
    else:
        raise pymol.CmdException("No helix state selected")

    if _self.count_atoms(f"name {atom_selection_name}", domain=tmp_connect):
        tmp_resv = iterate_to_list(tmp_connect,"resv")

        # Assign last res before adjustment is made
        last_resv = int(tmp_resv[0])

        if atom_selection_name == "O3'":
            tmp_resv[0] = str(tmp_resv[0] + 1)
        elif atom_selection_name == "P":
            tmp_resv[0] = str(tmp_resv[0] - 1)
        else:
            raise pymol.CmdException("Something went wrong with resv loop in extend_nuc")

        # Resv and resi assignment and testing
        resv = int(tmp_resv[0])
        resi = resv

        check_valid_attachment(nascent, atom_selection_name, selection, resv)

        reverse = False if atom_selection_name == "O3'" else True

        last_resi_sele = _prefix + "_last_resi"
        _self.select(last_resi_sele, f"(byobject {selection}) & chain {chain} & resi \\{last_resv}")

        if not nascent.dbl_helix:
            _self.alter(tmp_editor, f"chain='{chain}';segi='{chain}';resi=tmp_resv[0]",
                        space={'tmp_resv': tmp_resv})
            # Set parameters so that move_new_res can be used
            prev_oppo_res = "none"
            double_stranded_bool = False

            # Move and fuse
            move_new_res(frag_string, tmp_editor, last_resi_sele, prev_oppo_res, double_stranded_bool, nascent.form, antisense=reverse)
            _self.fuse(f"{tmp_editor} & name P", tmp_connect, mode=3)

            _self.select(tmp_domain, "byresi (pk1 | pk2)")

            if atom_selection_name == "O3'":
                _self.bond(f"{tmp_domain} & resi \\{last_resv} & name O3'",
                           f"{tmp_domain} & resi \\{resv} & name P")
            elif atom_selection_name == "P":
                _self.bond(f"{tmp_domain} & resi \\{resv} & name O3'",
                           f"{tmp_domain} & resi \\{last_resv} & name P")

        elif nascent.dbl_helix:
            #Initialize strings for selection
            prev_oppo_res = _prefix + "_prev_oppo_res"
            end_oppo_atom = _prefix + "_end_oppo_atom"

            #Initialize a double stranded bool for the detection of existing double strand
            double_stranded_bool = False
            base_pair_result = 0

            # Get oppo information
            atom_name_oppo = "O3'" if atom_selection_name == "P" else "P"

            # Get opposing chains and iterate looking for basepairs
            chains_oppo = get_chains_oppo(chain, tmp_connect)
            for tmp_chain_oppo in chains_oppo:
                # Create selection strings
                last_oppo_atom = _prefix + "_last_oppo_atom"
                first_oppo_atom = _prefix + "_first_oppo_atom"
                same_oppo_res =  _prefix + "_same_oppo_atom"

                _self.select(last_oppo_atom, f"last (({object} and chain {tmp_chain_oppo} and (not {tmp_editor})) and polymer.nuc)")
                _self.select(first_oppo_atom, f"first (({object} and chain {tmp_chain_oppo} and (not {tmp_editor})) and polymer.nuc)")
                base_pair_last = check_DNA_base_pair(last_oppo_atom, original_sele)
                base_pair_first = check_DNA_base_pair(first_oppo_atom, original_sele)

                same_base_flag = _self.select(same_oppo_res, f"({object} and (byres {last_oppo_atom}) and (byres {first_oppo_atom}))")

                if ((base_pair_first == 1 or base_pair_last == 1) and base_pair_result == 1 and same_base_flag == 0):
                    print("Multiple residues meet base pairing requirements. Building as if no opposing strand detected.")
                    base_pair_result = 0
                    break
                elif (base_pair_first == 1 and base_pair_last ==1):
                    chain_oppo = tmp_chain_oppo
                    base_pair_result = 1

                    check_DNA_base_pair(first_oppo_atom, original_sele)
                    atoms_first = iterate_to_list(_prefix + "_base_pair_dist","name")
                    check_DNA_base_pair(last_oppo_atom, original_sele)
                    atoms_last = iterate_to_list(_prefix + "_base_pair_dist", "name")

                    if (len(atoms_first) > len(atoms_last)):
                        _self.select(end_oppo_atom,first_oppo_atom)
                    else:
                        _self.select(end_oppo_atom,last_oppo_atom)

                elif (base_pair_last == 1):
                    base_pair_result = 1
                    chain_oppo = tmp_chain_oppo
                    _self.select(end_oppo_atom,last_oppo_atom)
                elif (base_pair_first == 1):
                    base_pair_result = 1
                    chain_oppo = tmp_chain_oppo
                    _self.select(end_oppo_atom,first_oppo_atom)
                else:
                    print("No based pair was found on chain ", tmp_chain_oppo)

            if (base_pair_result == 1):
                double_stranded_bool = True
                # Confirm base pair result and set _base_pair_dist
                check_DNA_base_pair(end_oppo_atom, original_sele)

                tmp_last_resv_oppo = iterate_to_list(end_oppo_atom,"resv")
                if len(tmp_last_resv_oppo) == 1:
                    last_resv_oppo = tmp_last_resv_oppo[0]
                else:
                    # This is arbitrary and may be changed. Using neg selected resv for now.
                    last_resv_oppo = -last_resv

                _self.select(prev_oppo_res, "byres (_tmp_editor_base_pair_dist)")
            elif (base_pair_result == 0):
                last_resv_oppo = -last_resv
                chain_oppo = get_new_chain(chain, tmp_connect)
            else:
                raise pymol.CmdException("Base pairing result is not returning 0 or 1")

            # Alter the opposing segi to match chain
            chain_oppo_sele = "_chain_oppo_sele"
            _self.select(chain_oppo_sele, f"chain {chain_oppo}")
            _self.alter(chain_oppo_sele,f"segi = '{chain_oppo}'")
            _self.delete(chain_oppo_sele)

            # Get oppo resv value based on selection
            if atom_selection_name == "O3'":
                resv_oppo =  last_resv_oppo - 1
            elif atom_selection_name == "P":
                resv_oppo = last_resv_oppo + 1

            # If picking O3', check O5' phosphate based on info found in this loop
            if (atom_selection_name == "O3'" and double_stranded_bool == True):
                tmp_phosphate_check = _prefix + "_phosphate_check"

                if(_self.select(tmp_phosphate_check, f"{prev_oppo_res} and name P ") == 0):
                    _self.select("pk1", f"{prev_oppo_res} and name O5'")
                    _self.select("pk2", f"{prev_oppo_res} and name O5'") # This is so attach frag has a pk2

                    attach_O5_phosphate()

                    _self.unpick()
                    _self.select("pk1",original_sele)
                    _self.select(tmp_connect,f"{selection}")
                    _self.select(prev_oppo_res, f"byres ({prev_oppo_res})")

                    if (_self.select(tmp_phosphate_check, f"{prev_oppo_res} and name P")==1):
                        print("Phosphate has been successfully added")

            # Move the fragment
            move_new_res(frag_string, tmp_editor, last_resi_sele, prev_oppo_res, double_stranded_bool, nascent.form, antisense=reverse)

            # Alter residues created
            _self.alter("_tmp_editor_newresA",f"chain='{chain}';segi='{chain}';resv={resv}")
            _self.alter("_tmp_editor_newresB",f"chain='{chain_oppo}';segi='{chain_oppo}';resv={resv_oppo}")

            # Connect the created residues
            if double_stranded_bool == True:
                bond_double_stranded(tmp_editor, object, chain, chain_oppo, resv, resv_oppo, last_resi_sele, prev_oppo_res,
                                     atom_selection_name, atom_name_oppo)
            elif double_stranded_bool == False:
                bond_single_stranded(tmp_editor, object, chain, resv, last_resi_sele, atom_selection_name, atom_name_oppo)
            else:
                raise pymol.CmdException("double_stranded_bool is not returning True or False")

        if nascent.nuc_type == "RNA":
            add2pO(tmp_domain, nascent.nuc_acid, resv)

    new_term = _prefix + "_new_selection"
    _self.select(new_term, f"{object} & chain {chain} & resi \\{resv} & name {atom_selection_name}")
    #_self.edit(new_term)

def fab(input,name=None,mode='peptide',resi=1,chain='',segi='',state=-1,
        dir=1,hydro=-1,ss=0,async_=0,quiet=1,_self=cmd, **kwargs):
    '''
DESCRIPTION

    Build a peptide

ARGUMENTS

    input = str: sequence in one-letter code

    name = str: name of object to create {default: }

    ss = int: Secondary structure 1=alpha helix, 2=antiparallel beta, 3=parallel beta, 4=flat

EXAMPLE

    fab ACDEFGH
    fab ACDEFGH, helix, ss=1
    '''
    async_ = int(kwargs.pop('async', async_))

    if kwargs:
        raise pymol.CmdException('unknown argument: ' + ', '.join(kwargs))

    if async_ < 1:
        r = _fab(input,name,mode,resi,chain,segi,
                 state,dir,hydro,ss,quiet,_self)
    else:
        fab_thread = threading.Thread(target=_fab, args=(input,name,mode,
                                                         resi,chain,
                                                         segi,state,dir,
                                                         hydro,ss,quiet,_self))
        fab_thread.setDaemon(1)
        fab_thread.start()
        r = DEFAULT_SUCCESS
    return r

def fnab(input, name=None, mode="DNA", form="B", dbl_helix=1, *, _self=cmd):
    """
DESCRIPTION

    Builds a nucleotide acid from sequence

    Fragments provided by:
    Lu, Xiang-Jun, Olson, Wilma K. 3DNA: a software package for the analysis,
    rebuilding and visualization of three-dimensional nucleic acid structures,
    Nucleic Acids Research, 32, W667-W675 (2004).

USAGE

    fnab input [, name [, type [, form [, dbl_helix ]]]]

ARGUMENTS

    input = str: Sequence as an array of one letter codes

    name = str: Name of the object to create {default: obj}

    mode = str: "DNA" or "RNA"

    form = str: "A" or "B"

    dbl_helix = bool (0/1): flag for using double helix in DNA

EXAMPLE

    fnab ATGCGATAC
    fnab ATGCGATAC, name=myDNA, mode=DNA, form=B, dbl_helix=1
    fnab AAUUUUCCG, mode=RNA
    """
    _self.unpick()

    if name is None:
        name = _self.get_unused_name(prefix="obj")
    elif name in _self.get_names('all'):
        name = _self.get_unused_name(prefix=name)

    dbl_helix = int(dbl_helix) > 0

    mode = mode.upper()
    if mode not in ('DNA', 'RNA'):
        raise pymol.CmdException("\"mode\" must be \"DNA\" or \"RNA\" only.")
    if mode == "RNA" and dbl_helix != 0:
        print ("Double helix RNA building is not currently supported.")
        dbl_helix = 0

    #first pass for error checking
    for oneNA in input:
        oneNA = oneNA.upper()
        threeNA = _oneNA_to_threeNA.get(oneNA)
        if threeNA is None:
            raise pymol.CmdException("\"%s\" not of %s type..." % (oneNA, mode))

        if threeNA not in _base_pair[mode]:
            raise pymol.CmdException("\"%s\" not of %s type..." % (oneNA, mode))

    for oneNA in input:
        oneNA = oneNA.upper()
        threeNA = _oneNA_to_threeNA[oneNA]
        attach_nuc_acid(selection="?pk1", nuc_acid=threeNA, nuc_type=mode,
                        object=name, form=form, dbl_helix=dbl_helix)
    _self.unpick()
    return DEFAULT_SUCCESS

def build_peptide(sequence,_self=cmd): # legacy
    for aa in sequence:
        attach_amino_acid("pk1",_aa_codes[aa])



'''
APBS wrapper

(c) 2012 Thomas Holder (https://github.com/speleo3/pymol-psico)
(c) 2016 Thomas Holder, Schrodinger, Inc.

License: BSD-2-Clause
'''

import os

from pymol import cmd, CmdException

template_apbs_in = '''
read
    mol pqr "{pqrfile}"
end
elec
    mg-auto
    mol 1

    fgcent {fgcent} # fine grid center
    cgcent mol 1    # coarse grid center
    fglen {fglen}
    cglen {cglen}
    dime {dime}
    lpbe          # l=linear, n=non-linear Poisson-Boltzmann equation
    bcfl sdh      # "Single Debye-Hueckel" boundary condition
    pdie 2.0      # protein dielectric
    sdie 78.0     # solvent dielectric
    chgm spl2     # Cubic B-spline discretization of point charges on grid
    srfm smol     # smoothed surface for dielectric and ion-accessibility coefficients
    swin 0.3
    temp 310.0    # temperature
    sdens 10.0
    calcenergy no
    calcforce no
    srad {srad}   # solvent radius

    ion charge +1 conc 0.15 radius 2.0
    ion charge -1 conc 0.15 radius 1.8

    write pot dx "{mapfile}"
end
quit
'''


def find_apbs_exe():
    import shutil
    exe = shutil.which('apbs')
    if not exe:
        exe = cmd.exp_path('$SCHRODINGER/utilities/apbs')
        if not os.path.exists(exe):
            return None
    return exe


def validate_apbs_exe(exe):
    '''Get and validate apbs executable.
    Raise CmdException if not found or broken.'''
    import subprocess

    if exe:
        exe = cmd.exp_path(exe)
    else:
        exe = find_apbs_exe() or 'apbs'

    try:
        r = subprocess.call([exe, "--version"],
                stdin=subprocess.PIPE,  # Windows pythonw fix
                stdout=open(os.devnull, "w"), stderr=subprocess.STDOUT)
        if r < 0:
            raise CmdException("Broken executable: " + exe)
    except OSError as e:
        print(e)
        raise CmdException("Cannot execute: " + exe)

    return exe

def map_new_apbs(name, selection='all', grid=0.5, buffer=10.0, state=1,
        preserve=0, exe='', assign=0, focus='', quiet=1, _template='',
        _proclist=None):
    '''
DESCRIPTION

    Create electrostatic potential map with APBS.

    "selection" needs partial charges (partial_charge) and radii (elec_radius)
    assigned. These can be loaded for example from a PQR file.

SEE ALSO

    map_new (coulomb), psico.electrostatics
    '''
    import tempfile, shutil, glob, subprocess

    selection = '(%s) and not solvent' % (selection)
    grid, buffer, state = float(grid), float(buffer), int(state)
    preserve, assign, quiet = int(preserve), int(assign), int(quiet)

    # path to apbs executable
    exe = validate_apbs_exe(exe)

    # temporary directory
    tempdir = tempfile.mkdtemp()
    if not quiet:
        print(' Tempdir:', tempdir)

    # filenames
    pqrfile = os.path.join(tempdir, 'mol.pqr')
    infile = os.path.join(tempdir, 'apbs.in')
    stem = os.path.join(tempdir, 'map')

    # temporary selection
    tmpname = cmd.get_unused_name('_sele')
    cmd.select(tmpname, selection, 0)

    cmd.save(pqrfile, tmpname, state, format='pqr', quiet=quiet)

    # grid dimensions
    extent = cmd.get_extent(tmpname, state)
    extentfocus = cmd.get_extent(focus) if focus else extent
    fglen = [(emax-emin + 2*buffer) for (emin, emax) in zip(*extentfocus)]
    cglen = [(emax-emin + 4*buffer) for (emin, emax) in zip(*extent)]
    # make coarse grid a cube (better for non-globular shapes)
    cglen = [max(cglen)] * 3

    cmd.delete(tmpname)

    apbs_in = {
        'pqrfile': pqrfile,
        'fgcent': 'mol 1',
        'fglen': '%f %f %f' % tuple(fglen),
        'cglen': '%f %f %f' % tuple(cglen),
        'srad': cmd.get('solvent_radius'),
        'mapfile': stem,
    }

    if focus:
        apbs_in['fgcent'] = '%f %f %f' % tuple((emax + emin) / 2.0
                for (emin, emax) in zip(*extentfocus))

    try:
        # apbs will fail if grid does not fit into memory
        # -> on failure repeat with larger grid spacing
        for _ in range(3):
            dime = [1 + max(64, n / grid) for n in fglen]
            apbs_in['dime'] = '%d %d %d' % tuple(dime)

            # apbs input file
            with open(infile, 'w') as f:
                f.write((_template or template_apbs_in).format(**apbs_in))

            # run apbs
            p = subprocess.Popen([exe, infile], cwd=tempdir)

            if _proclist is not None:
                _proclist.append(p)

            r = p.wait()

            if r == -15: # SIGTERM
                raise CmdException('apbs terminated')

            if r == 0:
                break

            if r in (-6, -9):
                grid *= 2.0
                if not quiet:
                    print(' Warning: retry with grid =', grid)
                continue

            raise CmdException('apbs failed with code ' + str(r))

        dx_list = glob.glob(stem + '*.dx')
        if not dx_list:
            dx_list = glob.glob(stem + '*.dxbin')
        if len(dx_list) != 1:
            raise CmdException('dx file missing')

        # load map
        cmd.load(dx_list[0], name, quiet=quiet)
    except OSError as e:
        print(e)
        raise CmdException('Cannot execute "%s"' % (exe))
    finally:
        if not preserve:
            shutil.rmtree(tempdir)
        elif not quiet:
            print(' Notice: not deleting %s' % (tempdir))


# vi: ts=4:sw=4:smarttab:expandtab
